from typing import Optional, List, Dict, Any, Tuple
from dataclasses import dataclass, field
import shutil
import os
import sys
import base64
import fcntl
import shlex
import logging
import subprocess
import time

from qemu_compose.qemu.machine import QEMUMachine
from qemu_compose.local_store import LocalStore
from qemu_compose.instance import prepare_ssh_key
from qemu_compose.utils.hostnames import to_valid_hostname
from qemu_compose.utils.vsock import get_available_guest_cid
from qemu_compose.utils import StreamWrapper
from qemu_compose.image import ImageManifest, load_image_by_id, load_image_by_name, DiskSpec

from .name import check_and_get_name
from .http import HttpServer
from .terminal import Terminal
from . import new_random_vmid


logger = logging.getLogger("qemu-compose.instance.qemu_runner")


def create_overlay(base_path: str, base_format: str, overlay_path: str) -> int:
    cmd = [
        "qemu-img", "create",
        "-b", base_path,
        "-F", base_format,
        "-f", "qcow2",
        overlay_path,
    ]
    try:
        res = subprocess.run(cmd, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if res.returncode != 0:
            print(res.stderr, file=sys.stderr, flush=True)
        return res.returncode
    except FileNotFoundError:
        print("Error: 'qemu-img' binary not found in PATH", file=sys.stderr, flush=True)
        return 127

def drive_param_for(overlay_path: str, spec: DiskSpec) -> str:
    # Build a '-drive' parameter string combining manifest opts with required pieces.
    opts = []
    # Always use qcow2 overlay per requirement
    opts.append(f"file={overlay_path}")
    if spec.format:
        opts.append("format=" + spec.format)
    if spec.opts:
        opts.append(spec.opts)
    return ",".join(opts)


def extract_format_or_default(mapping: Optional[dict], key: str, env: dict, default=None):
    value = mapping.get(key) if mapping else key
    if value:
        # FIXME: format has security issues
        return str(value).format(**env)
    return default


@dataclass(frozen=True)
class HttpServeConfig:
    listen: str
    port: int
    root: str
    access_ip: Optional[str] = None

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "HttpServeConfig":
        return cls(
            listen=d.get("listen", "127.0.0.1"),
            port=int(d.get("port", 8888)),
            root=d.get("root", ""),
            access_ip=d.get("access_ip")
        )

@dataclass(frozen=True)
class QemuConfig:
    name: Optional[str] = None
    binary: Optional[str] = None
    network: Optional[str] = None     # could be "none", "user", etc, default set to "user" when left None
    image: Optional[str] = None
    env: Dict[str, str] = field(default_factory=dict)
    qemu_args: List[Dict[str, str]] = field(default_factory=list)
    ports: List[str] = field(default_factory=list)
    volumes: List[str] = field(default_factory=list)
    boot_commands: List[Dict[str, Any]] = field(default_factory=list)
    before_script: List[str] = field(default_factory=list)
    after_script: List[str] = field(default_factory=list)
    http_serve: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "QemuConfig":
        return cls(
            name=d.get("name"),
            binary=d.get("binary"),
            network=d.get("network"),
            image=d.get("image"),
            env=d.get("env", []),
            qemu_args=d.get("qemu_args", []),
            ports=d.get("ports", []),
            volumes=d.get("volumes", []),
            boot_commands=d.get("boot_commands", []),
            before_script=d.get("before_script", []),
            after_script=d.get("after_script", []),
            http_serve=d.get("http_serve", {}),
        )


class QemuRunner(QEMUMachine):
    def __init__(self, config: QemuConfig, store: LocalStore, cwd: str):
        self.config = config
        self.store = store
        self.cwd = cwd

        self.vm_name: Optional[str] = None
        self.lock_fd: Optional[int] = None
        self.cid: Optional[int] = None
        self.vmid: Optional[str] = None
        self.log_file = None
        self.image_manifest: Optional[ImageManifest] = None
        self.storage_overlays: List[Tuple[str, DiskSpec]] = []
        self.virtiofs_children: List[subprocess.Popen] = []

        if config.binary:
            binary = config.binary
        else:
            binary = shutil.which('qemu-system-x86_64')

        if not binary:
            raise FileNotFoundError("QEMU binary not found")
        
        super().__init__(binary=binary)

        self.add_monitor_null()
        self.set_qmp_monitor(True)
        self.set_console(console_chardev='socket', device_type='isa-serial')

    @property
    def instance_dir(self) -> str:
        if self.vmid is None:
            raise ValueError("vmid is not set")
        return self.store.instance_dir(self.vmid)

    def check_and_lock(self) -> int:
        if self.config.image is not None:
            manifest = load_image_by_id(self.store.image_root, self.config.image)

            if manifest is None:
                manifest = load_image_by_name(self.store.image_root, self.config.image)

            if manifest is None:
                print(f"Image '{self.config.image}' not found in local store", file=sys.stderr)
                return 126
            self.image_manifest = manifest

        try:
            self.vm_name = check_and_get_name(self.store.instance_root, self.config.name)
        except ValueError as e:
            print(str(e), file=sys.stderr)
            return 125

        self.cid = get_available_guest_cid(1000)
        if self.cid is None:
            print("no available guest cid found, please make sure vhost_vsock module loaded", file=sys.stderr)
            return 124

        self.vmid = new_random_vmid(self.store.instance_root)

        log_path = os.path.join(self.store.instance_dir(self.vmid), "qemu-compose.log")
        self.log_file = open(log_path, "wb")
        logging.basicConfig(level=logging.INFO, stream=StreamWrapper(self.log_file))

        try:
            instance_dir = self.store.instance_dir(self.vmid)
        except OSError as e:
            print(f"Failed to create instance dir {self.vmid}: {e}", file=sys.stderr)
            return 123
        
        try:
            # Acquire exclusive lock on instance_dir before any launch
            # lock early to prevent prune procedure removing contents before qemu starts
            flags = os.O_RDONLY
            if hasattr(os, 'O_DIRECTORY'):
                flags |= os.O_DIRECTORY
            self.lock_fd = os.open(instance_dir, flags)
            fcntl.flock(self.lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except BlockingIOError:
            print(f"Failed to lock instance dir {instance_dir}", file=sys.stderr)
            return 122
        
        return 0

    def prepare_env(self, env_update: Optional[Dict[str, str]] = None):
        term_size = os.get_terminal_size()

        env = {
            'CWD': self.cwd,
            'GATEWAY_IP': '10.0.2.2',   # qemu user network default gateway ip
            'TERM_ROWS': term_size.lines,
            'TERM_COLS': term_size.columns,
            'ID': self.vmid,
            'STORAGE_PATH': self.store.data_dir,
            'IMAGE_ROOT': self.store.image_root,
            'INSTANCE_ROOT': self.store.instance_root,
            'INSTANCE_DIR': self.instance_dir,
        }

        if self.config.image:
            env['IMAGE_TAG'] = self.config.image

        if self.image_manifest is not None:
            env["IMAGE_DIR"] = os.path.join(self.store.image_root, self.image_manifest.id)
            env["IMAGE_ID"] = self.image_manifest.id

        if self.config.env:
            for k in self.config.env:
                env[k] = self.config.env[k]

        logger.info("change directory to %s" % env['CWD'])
        os.chdir(env['CWD'])
        
        http_port = None
        if self.config.http_serve:
            http_serve_config: dict = self.config.http_serve
            http_listen = extract_format_or_default(http_serve_config, 'listen', env, default='0.0.0.0')
            http_port = int(extract_format_or_default(http_serve_config, 'port', env, default=8888))
            http_root = extract_format_or_default(http_serve_config, 'root', env, default=env['CWD'])

            http_server = HttpServer(http_listen, http_port, http_root)
            http_server.start()
            logger.info('HTTP server started on %s:%d, serving %s' % (http_listen, http_port, http_root))

        if http_port is not None:
            env['HTTP_PORT'] = http_port
            access_ip = extract_format_or_default(self.config.http_serve, 'access_ip', env, default=env['GATEWAY_IP']) if self.config.http_serve else env['GATEWAY_IP']
            env['HTTP_HOST'] = access_ip

        if env_update:
            env.update(env_update)

        self.env = env

    def prepare_storage(self):
        if self.image_manifest is None:
            return 0
        
        image_dir = os.path.join(self.store.image_root, self.image_manifest.id)

        self.storage_overlays = []

        for disk_spec in self.image_manifest.disks:
            base_disk_path = os.path.join(image_dir, disk_spec.filename)
            overlay_path = os.path.join(self.instance_dir, disk_spec.filename)
            rc = create_overlay(base_disk_path, disk_spec.format, overlay_path)
            if rc != 0:
                print(f"Failed to create overlay for disk {disk_spec.filename}", file=sys.stderr, flush=True)
                return rc
            self.storage_overlays.append((overlay_path, disk_spec))

        return 0

    def execute_script(self, script_key: str):
        script_target = getattr(self.config, script_key, None)

        if script_target:
            for line in script_target:
                command = extract_format_or_default(None, line, self.env)
                if command:
                    subprocess.run(command.strip(), shell=True, check=True)

    def setup_qemu_args(self):
        # the very default args

        vm_mem_size = "1G"

        default_args = {
            'cpu': 'max',
            'machine': 'type=q35,hpet=off',
            'accel': 'kvm',
            'm': vm_mem_size,
            'smp': str(os.cpu_count()),
        }

        # image provided args override our defaults
        if self.image_manifest is not None and self.image_manifest.qemu_args:
            for i, a in enumerate(self.image_manifest.qemu_args):
                if not a.startswith('-'):
                    continue
                if a[1:] in default_args:
                    default_args.pop(a[1:], None)

                    if a[1:] == "m":
                        if i + 1 < len(self.image_manifest.qemu_args):
                            vm_mem_size = self.image_manifest.qemu_args[i + 1]


        # user provided args override image defaults
        for block in self.config.qemu_args:
            for key in block:
                val = extract_format_or_default(block, key, self.env)
                if key in default_args and isinstance(val, str):
                    default_args[key] = val

                    if key == "m":
                        vm_mem_size = val

        args = []

        # vm name first
        if self.vm_name:
            args.append('-name')
            args.append(self.vm_name)

        # then our safe defaults
        for key in default_args:
            args.append('-' + key)
            args.append(default_args[key])

        # then network setup

        hostname = None
        if self.vm_name:
            hostname = to_valid_hostname(self.vm_name)

        if hostname:
            # https://systemd.io/CREDENTIALS/
            args.append('-smbios')
            args.append('type=11,value=io.systemd.credential:system.hostname=' + hostname)

        def parse_port_spec(spec: str) -> Optional[Tuple[str, str, str, str]]:
            # Support forms:
            #  - host_ip:host_port:vm_port
            #  - host_port:vm_port
            # Each of the above may be suffixed with "/tcp" or "/udp"
            proto = 'tcp'
            body, sep, suffix = spec.partition('/')
            if sep:
                proto = suffix.strip().lower() or 'tcp'
                if proto not in ('tcp', 'udp'):
                    proto = 'tcp'
            parts = [p.strip() for p in body.split(':')]
            if len(parts) == 3:
                return proto, parts[0], parts[1], parts[2]
            if len(parts) == 2:
                return proto, '', parts[0], parts[1]
            return None

        def format_hostfwd(item: Tuple[str, str, str, str]) -> str:
            proto, host_ip, host_port, vm_port = item
            return f",hostfwd={proto}:{host_ip}:{host_port}-:{vm_port}"

        def hostfwd_segments(ports: List[str]) -> str:
            return ''.join(
                map(
                    format_hostfwd,
                    filter(None, map(parse_port_spec, ports))
                )
            )

        # Bind-mount style volumes implemented via virtio-fs. Spec format:
        #   src:dst[:ro]
        # Examples:
        #   /host/path:/mnt/data
        #   /host/path:/mnt/readonly:ro
        def parse_volume_spec(spec: str) -> Optional[Tuple[str, str, bool]]:
            parts = [p.strip() for p in spec.split(':')]
            if len(parts) < 2:
                return None
            src = parts[0]
            dst = parts[1]
            ro = any(p.lower() == 'ro' for p in parts[2:]) if len(parts) > 2 else False
            if not src or not dst:
                return None
            return src, dst, ro

        def volume_tag_for(dst: str, idx: int) -> str:
            base = os.path.basename(dst) or f"vol{idx}"
            sanitized = ''.join(ch if ch.isalnum() or ch in ('-', '_') else '_' for ch in base)
            return f"{sanitized}-{idx}"

        def start_virtiofsd(shared_dir: str, socket_path: str, read_only: bool) -> Optional[subprocess.Popen]:
            unshare_bin = shutil.which('unshare')

            virtiofsd_bin = shutil.which('virtiofsd', path="/usr/lib:/usr/libexec")
            if virtiofsd_bin is None or unshare_bin is None:
                logger.warning("virtiofsd or unshare command not found; volume '%s' will not be available", shared_dir)
                return None
            # Prefer running virtiofsd under unshare with userns mapping when available
            if unshare_bin is not None:
                cmd = [
                    unshare_bin,
                    '-r', '--map-auto', '--',
                    virtiofsd_bin,
                    '--shared-dir', shared_dir,
                    '--socket-path', socket_path,
                    '--cache', 'never',
                    '--allow-direct-io',
                    '--allow-mmap',
                    '--thread-pool-size', '8',
                    '--sandbox', 'chroot',
                ]
            else:
                cmd = [
                    virtiofsd_bin,
                    '--shared-dir', shared_dir,
                    '--socket-path', socket_path,
                    '--cache', 'never',
                    '--allow-direct-io',
                    '--allow-mmap',
                    '--thread-pool-size', '8',
                    '--sandbox', 'chroot',
                ]
            if read_only:
                cmd.append('--readonly')
            try:
                logger.info("running virtiofsd %s" % (" ".join(shlex.quote(p) for p in cmd), ))
                proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                # proc = subprocess.Popen(["/usr/bin/tail", "-f", "/dev/null"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                return proc
            except Exception as e:
                logger.warning("failed to start virtiofsd for %s: %s", shared_dir, e)
                return None

        def drain_proc_output(proc: subprocess.Popen):
            for s in (proc.stdout, proc.stderr):
                if s is None:
                    continue
                if not s.readable():
                    continue

                os.set_blocking(s.fileno(), False)
                try:
                    line:bytes = s.read(1024)
                    if line:
                        logger.debug("virtiofsd: %s", line.decode('utf-8', errors='replace').rstrip())
                except Exception:
                    pass

        def wait_for_socket(proc: subprocess.Popen, path: str, timeout_sec: float = 3.0, interval_sec: float = 0.05) -> bool:
            deadline = time.time() + timeout_sec
            drain_proc_output(proc)
            while time.time() < deadline:
                if os.path.exists(path):
                    return True
                time.sleep(interval_sec)
                drain_proc_output(proc)
            return os.path.exists(path)

        if self.config.network is None or self.config.network.lower() == 'user':
            # add user network
            args.append('-netdev')
            # https://man.archlinux.org/man/qemu.1.en#hostname=name
            base = 'user,id=user.qemu-compose%s' % (',hostname=' + hostname if hostname else '',)
            netdev_opts = base + hostfwd_segments(self.config.ports or [])
            args.append(netdev_opts)
            args.append('-device')
            args.append('virtio-net,netdev=user.qemu-compose')

        if self.cid:
            args.append("-device")
            args.append("vhost-vsock-pci,id=vhost-vsock-pci0,guest-cid=%d" % self.cid)

        assert self.vmid is not None
        pub_bytes = prepare_ssh_key(self.instance_dir, self.vmid)
        pub_b64 = base64.b64encode(pub_bytes).decode('ascii')

        args.append('-smbios')
        args.append(f'type=11,value=io.systemd.credential.binary:ssh.authorized_keys.root={pub_b64}')

        # storage disks
        for overlay_path, spec in self.storage_overlays:
            drive_param = drive_param_for(overlay_path, spec)
            args.append('-drive')
            args.append(drive_param)

        # volumes via virtio-fs and fstab entries
        fstab_entries: List[str] = []
        for i, vol_spec in enumerate(self.config.volumes or []):
            parsed = parse_volume_spec(vol_spec)
            if not parsed:
                continue
            src, dst, ro = parsed
            tag = volume_tag_for(dst, i)
            socket_path = os.path.join(self.instance_dir, f"virtiofs-{tag}.sock")
            child = start_virtiofsd(src, socket_path, ro)
            if child is None:
                continue

            # Wait for server socket to exist before wiring chardev, to avoid QEMU connect errors
            if not wait_for_socket(child, socket_path, 300):
                logger.warning("virtiofsd socket not ready, skipping mount %s -> %s", src, dst)
                # Terminate child since we won't use it
                try:
                    child.terminate()
                except Exception:
                    pass
                continue

            self.virtiofs_children.append(child)
            args.append('-chardev')
            args.append(f"socket,id=qcfs-char{i},path={socket_path}")
            args.append('-device')
            args.append(f"vhost-user-fs-pci,chardev=qcfs-char{i},tag={tag}")
            ro_suffix = ',ro' if ro else ''
            fstab_entries.append(f"{tag} {dst} virtiofs defaults{ro_suffix} 0 0")

        if fstab_entries:
            # Force use of memory sharing with virtiofsd
            # see https://github.com/virtio-win/kvm-guest-drivers-windows/wiki/Virtiofs:-Shared-file-system

            args.extend([
                "-object",
                "memory-backend-file,id=qc-mem,size=%s,mem-path=/dev/shm,share=on" % vm_mem_size,
                "-numa",
                "node,memdev=qc-mem",
            ])

            try:
                fstab_str = "\n".join(fstab_entries)
                fstab_b64 = base64.b64encode(fstab_str.encode('utf-8')).decode('ascii')
                args.append('-smbios')
                args.append(f'type=11,value=io.systemd.credential.binary:fstab.extra={fstab_b64}')
            except Exception as e:
                logger.warning("failed to encode fstab entries: %s", e)

        # image provided args append after defaults
        if self.image_manifest is not None and self.image_manifest.qemu_args:
            for arg in self.image_manifest.qemu_args:
                val = extract_format_or_default(None, arg, self.env)
                args.append(val)

        # user provided args append after defaults
        for block in self.config.qemu_args:
            for key in block:
                val = extract_format_or_default(block, key, self.env)
                if key in default_args:
                    continue
                args.append('-' + key)
                if val is not None:
                    args.append(val)

        self.add_args(*args)

    def start(self):
        self.launch()

        self.term = Terminal(self.console_file, self.log_file)

        try:
            pid = self.get_pid()
        except Exception:
            pid = None
        try:
            with open(os.path.join(self.instance_dir, "qemu.pid"), "w") as f:
                f.write("%s" % (str(pid) if pid is not None else ""))
            with open(os.path.join(self.instance_dir, "cid"), "w") as f:
                f.write(str(self.cid))
            with open(os.path.join(self.instance_dir, "name"), "w") as f:
                f.write(str(self.vm_name) if self.vm_name is not None else "")
            with open(os.path.join(self.instance_dir, "instance-id"), "w") as f:
                f.write(str(self.vmid))
        except Exception as e:
            logger.warning("failed to write instance metadata: %s", e)

    def interact(self):
        boot_commands = self.config.boot_commands
        if boot_commands:
            self.term.run_batch(boot_commands, env_variables=self.env)
        else:
            self.term.interact(raw_mode=True)

    def cleanup(self):
        self._load_io_log()
        logger.info('vm.process_io_log = %r' % (self.get_log(), ))

        try:
            if self.lock_fd is not None:
                fcntl.flock(self.lock_fd, fcntl.LOCK_UN)
                os.close(self.lock_fd)
        except Exception as e:
            logger.warning("failed to unlock instance dir: %s", e)

        # Ensure virtiofsd processes are cleaned up
        try:
            children = getattr(self, 'virtiofs_children', None)
            if children:
                # Ask nicely first
                for p in children:
                    try:
                        if p and p.poll() is None:
                            p.terminate()
                    except Exception as e:
                        logger.debug("terminate virtiofsd failed: %s", e)
                # Wait briefly
                for p in children:
                    try:
                        if p and p.poll() is None:
                            p.wait(timeout=2)
                    except subprocess.TimeoutExpired:
                        try:
                            p.kill()
                        except Exception as e:
                            logger.debug("kill virtiofsd failed: %s", e)
                    except Exception as e:
                        logger.debug("wait virtiofsd failed: %s", e)
                # Close stdio pipes
                for p in children:
                    for stream_name in ("stdin", "stdout", "stderr"):
                        try:
                            s = getattr(p, stream_name, None)
                            if s:
                                s.close()
                        except Exception:
                            pass
                self.virtiofs_children = []
        except Exception as e:
            logger.warning("failed to cleanup virtiofsd: %s", e)



from typing import List, Optional, Tuple
import os

from qemu_compose.utils.human_readable import humanize_age, human_readable_size
from qemu_compose.utils import list_subdirs

from .manifest import ImageManifest, RepoTag, DiskSpec



def _short_image_id(digest: Optional[str]) -> str:
    if not digest:
        return "<none>"
    if digest.startswith("sha256:"):
        return digest.split(":", 1)[1][:12]
    # Fallback: first 12 chars
    return digest[:12]

def _file_size(path: str) -> int:
    try:
        return os.path.getsize(path)
    except OSError:
        return 0


def _size_from_manifest(image_dir: str, manifest: ImageManifest) -> int:
    disks = manifest.disks
    if isinstance(disks, list):
        return sum(_file_size(os.path.join(image_dir, f.filename)) for f in disks if isinstance(f, DiskSpec))
    return 0

def _rows_for_image(image_root: str, image_id: str) -> List[Tuple[str, str, str, str, str]]:
    dir_path = os.path.join(image_root, image_id)

    manifest = ImageManifest.load_file(dir_path)

    created_human = humanize_age(manifest.created)
    image_id_short = _short_image_id(manifest.digest)

    size_bytes = _size_from_manifest(dir_path, manifest)
    size_human = human_readable_size(size_bytes)

    def to_row(repo_tag: RepoTag) -> Tuple[str, str, str, str, str]:
        return (repo_tag.repo or "<none>", repo_tag.tag or "<none>", image_id_short, created_human, size_human)

    return [to_row(t) for t in manifest.repo_tags]


def list_image(image_root: str) -> List[Tuple[str, str, str, str, str]]:
    return [row for image_id in list_subdirs(image_root) for row in _rows_for_image(image_root, image_id)]

def load_image_by_id(image_root: str, image_id: str) -> Optional[ImageManifest]:
    dir_path = os.path.join(image_root, image_id)
    if not os.path.exists(dir_path) or not os.path.isdir(dir_path):
        return None
    return ImageManifest.load_file(dir_path)

def load_image_by_name(image_root: str, name: str) -> Optional[ImageManifest]:
    for image_id in list_subdirs(image_root):
        dir_path = os.path.join(image_root, image_id)
        if not os.path.exists(dir_path) or not os.path.isdir(dir_path):
            continue

        manifest = ImageManifest.load_file(dir_path)

        if manifest.has_repo_tag(name):
            return manifest
    return None

def resolve_image_by_prefix(image_root: str, token: str) -> Tuple[Optional[str], List[str]]:
    ids = list_subdirs(image_root)
    if token in ids:
        return token, [token]

    matches = [i for i in ids if i.startswith(token)]
    if len(matches) == 1:
        return matches[0], matches

    return None, matches

def resolve_image(image_root: str, token: str):
    if (found := load_image_by_name(image_root, token)):
        return found.id, [found.id]
    
    return resolve_image_by_prefix(image_root, token)



from __future__ import annotations
from typing import Optional, List

import os
import sys
import logging

from qemu_compose.local_store import LocalStore
from qemu_compose.image import resolve_image
from qemu_compose.instance.qemu_runner import QemuRunner, QemuConfig
from qemu_compose.qemu.machine.machine import AbnormalShutdown

logger = logging.getLogger("qemu-compose.cmd.run_command")

def command_run(*, image_hint: str, name: Optional[str], publish: Optional[List[str]] = None, volumes: Optional[List[str]] = None) -> int:
    store = LocalStore()

    resolved_id, prefix_matches = resolve_image(store.image_root, image_hint)

    # Resolve image id: exact, unique prefix, or repo_tag
    if resolved_id is None:
        if prefix_matches:
            preview = ", ".join(sorted(set(prefix_matches))[:8])
            more = "" if len(set(prefix_matches)) <= 8 else f" ... and {len(set(prefix_matches))-8} more"
            print(f"Error: image identifier '{image_hint}' is ambiguous; matches: {preview}{more}", flush=True, file=sys.stderr)
        else:
            print(f"Error: image not found: {image_hint}", flush=True, file=sys.stderr)
        return 1

    cwd = os.getcwd()

    config = QemuConfig(name=name, image=resolved_id, ports=list(publish or []), volumes=list(volumes or []))
    vm = QemuRunner(config, store, cwd)

    if (exit_code := vm.check_and_lock()) > 0:
        return exit_code

    vm.prepare_env()

    if (exit_code := vm.prepare_storage()) > 0:
        return exit_code

    vm.execute_script('before_script')
    vm.setup_qemu_args()

    try:
        vm.start()
        vm.interact()
        vm.execute_script('after_script')
    except KeyboardInterrupt:
        logger.warning("Keyboard interrupt, shutting down vm...")
    finally:
        try:
            if vm is not None and vm.is_running():
                vm.shutdown(hard=True)
        except AbnormalShutdown:
            logger.error('abnormal shutdown exception')
        finally:
            if vm is not None:
                vm.cleanup()
    return 0



import os
import fcntl
import struct
import errno
import logging

logger = logging.getLogger(__name__)

VHOST_VSOCK_SET_GUEST_CID = 0x4008AF60

VSOCK_PATH = '/dev/vhost-vsock'


def get_available_guest_cid(start_guest_cid: int = 1000) -> None | int:
    """
    get available guest cid

    :param guest_cid: start_guest_cid
    :return: available guest cid if success, else None
    """
    try:
        vsock_fd = os.open(VSOCK_PATH, os.O_RDWR)
    except FileNotFoundError:
        logger.info(f"could not open {VSOCK_PATH}, make sure vhost_vsock module loaded by running `modprobe vhost_vsock` using root")
        return None
    except PermissionError:
        logger.info(f"could not open {VSOCK_PATH}, please run as root")
        return None

    guest_cid = start_guest_cid
    try:
        while guest_cid <= 0xFFFFFFFF:  # U32_MAX
            cid_c = struct.pack('L', guest_cid)
            try:
                fcntl.ioctl(vsock_fd, VHOST_VSOCK_SET_GUEST_CID, cid_c)
            except IOError as e:
                if e.errno == errno.EADDRINUSE:
                    guest_cid += 1
                    continue
                else:
                    raise e
            else:
                return guest_cid
        return None
    finally:
        os.close(vsock_fd)



"""
QEMU Console Socket Module:

This python module implements a ConsoleSocket object,
which can drain a socket and optionally dump the bytes to file.
"""
# Copyright 2020 Linaro
#
# Authors:
#  Robert Foley <robert.foley@linaro.org>
#
# This code is licensed under the GPL version 2 or later.  See
# the COPYING file in the top-level directory.
#

from collections import deque
import socket
import threading
import time
from typing import Deque, Optional


class ConsoleSocket(socket.socket):
    """
    ConsoleSocket represents a socket attached to a char device.

    :param address: An AF_UNIX path or address.
    :param sock_fd: Optionally, an existing socket file descriptor.
                    One of address or sock_fd must be specified.
    :param file: Optionally, a filename to log to.
    :param drain: Optionally, drains the socket and places the bytes
                  into an in memory buffer for later processing.
    """
    def __init__(self,
                 address: Optional[str] = None,
                 sock_fd: Optional[int] = None,
                 file: Optional[str] = None,
                 drain: bool = False):
        if address is None and sock_fd is None:
            raise ValueError("one of 'address' or 'sock_fd' must be specified")
        if address is not None and sock_fd is not None:
            raise ValueError("can't specify both 'address' and 'sock_fd'")

        self._recv_timeout_sec = 300.0
        self._sleep_time = 0.5
        self._buffer: Deque[int] = deque()
        if address is not None:
            socket.socket.__init__(self, socket.AF_UNIX, socket.SOCK_STREAM)
            self.connect(address)
        else:
            assert sock_fd is not None
            socket.socket.__init__(self, fileno=sock_fd)
        self._logfile = None
        if file:
            # pylint: disable=consider-using-with
            self._logfile = open(file, "bw")
        self._open = True
        self._drain_thread = None
        if drain:
            self._drain_thread = self._thread_start()

    def __repr__(self) -> str:
        tmp = super().__repr__()
        tmp = tmp.rstrip(">")
        tmp = "%s,  logfile=%s, drain_thread=%s>" % (tmp, self._logfile,
                                                     self._drain_thread)
        return tmp

    def _drain_fn(self) -> None:
        """Drains the socket and runs while the socket is open."""
        while self._open:
            try:
                self._drain_socket()
            except socket.timeout:
                # The socket is expected to timeout since we set a
                # short timeout to allow the thread to exit when
                # self._open is set to False.
                time.sleep(self._sleep_time)

    def _thread_start(self) -> threading.Thread:
        """Kick off a thread to drain the socket."""
        # Configure socket to not block and timeout.
        # This allows our drain thread to not block
        # on receive and exit smoothly.
        socket.socket.setblocking(self, False)
        socket.socket.settimeout(self, 1)
        drain_thread = threading.Thread(target=self._drain_fn)
        drain_thread.daemon = True
        drain_thread.start()
        return drain_thread

    def close(self) -> None:
        """Close the base object and wait for the thread to terminate"""
        if self._open:
            self._open = False
            if self._drain_thread is not None:
                thread, self._drain_thread = self._drain_thread, None
                thread.join()
            socket.socket.close(self)
            if self._logfile:
                self._logfile.close()
                self._logfile = None

    def _drain_socket(self) -> None:
        """process arriving characters into in memory _buffer"""
        data = socket.socket.recv(self, 1)
        if self._logfile:
            self._logfile.write(data)
            self._logfile.flush()
        self._buffer.extend(data)

    def recv(self, bufsize: int = 1, flags: int = 0) -> bytes:
        """Return chars from in memory buffer.
           Maintains the same API as socket.socket.recv.
        """
        if self._drain_thread is None:
            # Not buffering the socket, pass thru to socket.
            return socket.socket.recv(self, bufsize, flags)
        assert not flags, "Cannot pass flags to recv() in drained mode"
        start_time = time.time()
        while len(self._buffer) < bufsize:
            time.sleep(self._sleep_time)
            elapsed_sec = time.time() - start_time
            if elapsed_sec > self._recv_timeout_sec:
                raise socket.timeout
        return bytes((self._buffer.popleft() for i in range(bufsize)))

    def setblocking(self, value: bool) -> None:
        """When not draining we pass thru to the socket,
           since when draining we control socket blocking.
        """
        if self._drain_thread is None:
            socket.socket.setblocking(self, value)

    def settimeout(self, value: Optional[float]) -> None:
        """When not draining we pass thru to the socket,
           since when draining we control the timeout.
        """
        if value is not None:
            self._recv_timeout_sec = value
        if self._drain_thread is None:
            socket.socket.settimeout(self, value)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Iterable, List, Optional

from qemu_compose.local_store import LocalStore

@dataclass(frozen=True)
class InstanceMeta:
    instance_id: str
    name: Optional[str]
    cid: Optional[int]
    pid: Optional[int]


def _safe_read(path: str) -> Optional[str]:
    try:
        with open(path, "r") as f:
            return f.read().strip() or None
    except Exception:
        return None


def _to_int(s: Optional[str]) -> Optional[int]:
    try:
        return int(s) if s is not None else None
    except Exception:
        return None


def _is_pid_running(pid: Optional[int]) -> bool:
    if pid is None:
        return False
    try:
        os.kill(pid, 0)
        return True
    except Exception:
        return False


def _list_instance_ids(store: LocalStore) -> List[str]:
    try:
        return sorted(
            [d for d in os.listdir(store.instance_root) if os.path.isdir(os.path.join(store.instance_root, d))]
        )
    except FileNotFoundError:
        return []


def _read_instance_meta(store: LocalStore, instance_id: str) -> InstanceMeta:
    # Avoid side effects: do not create directories while reading
    base = os.path.join(store.instance_root, instance_id)
    name = _safe_read(os.path.join(base, "name"))
    cid = _to_int(_safe_read(os.path.join(base, "cid")))
    pid = _to_int(_safe_read(os.path.join(base, "qemu.pid")))
    return InstanceMeta(instance_id=instance_id, name=name, cid=cid, pid=pid)


def _collect_instances(store: LocalStore) -> List[InstanceMeta]:
    return [_read_instance_meta(store, iid) for iid in _list_instance_ids(store)]


def _filter_instances(instances: Iterable[InstanceMeta], show_all: bool) -> List[InstanceMeta]:
    return [m for m in instances if show_all or _is_pid_running(m.pid)]


def _format_row(meta: InstanceMeta) -> str:
    status = "running" if _is_pid_running(meta.pid) else "exited"
    name = meta.name or "-"
    cid = str(meta.cid) if meta.cid is not None else "-"
    pid = str(meta.pid) if meta.pid is not None else "-"
    return f"{meta.instance_id:12}  {name:20}  {cid:6}  {pid:8}  {status}"


def _print_table(instances: Iterable[InstanceMeta]) -> None:
    header = f"{'INSTANCE_ID':12}  {'NAME':20}  {'CID':6}  {'QEMU PID':8}  STATUS"
    print(header)
    print("-" * len(header))
    for m in instances:
        print(_format_row(m))


def command_ps(show_all: bool) -> int:
    store = LocalStore()
    instances = _collect_instances(store)
    filtered = _filter_instances(instances, show_all)
    _print_table(filtered)
    return 0



"""
QEMU Object Model FUSE filesystem tool

This script offers a simple FUSE filesystem within which the QOM tree
may be browsed, queried and edited using traditional shell tooling.

This script requires the 'fusepy' python package.


usage: qom-fuse [-h] [--socket SOCKET] <mount>

Mount a QOM tree as a FUSE filesystem

positional arguments:
  <mount>               Mount point

optional arguments:
  -h, --help            show this help message and exit
  --socket SOCKET, -s SOCKET
                        QMP socket path or address (addr:port). May also be
                        set via QMP_SOCKET environment variable.
"""
##
# Copyright IBM, Corp. 2012
# Copyright (C) 2020 Red Hat, Inc.
#
# Authors:
#  Anthony Liguori   <aliguori@us.ibm.com>
#  Markus Armbruster <armbru@redhat.com>
#
# This work is licensed under the terms of the GNU GPL, version 2 or later.
# See the COPYING file in the top-level directory.
##

import argparse
from errno import ENOENT, EPERM
import stat
import sys
from typing import (
    IO,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Union,
)

import fuse
from fuse import FUSE, FuseOSError, Operations

from qemu.qmp import ExecuteError

from .qom_common import QOMCommand


fuse.fuse_python_api = (0, 2)


class QOMFuse(QOMCommand, Operations):
    """
    QOMFuse implements both fuse.Operations and QOMCommand.

    Operations implements the FS, and QOMCommand implements the CLI command.
    """
    name = 'fuse'
    help = 'Mount a QOM tree as a FUSE filesystem'
    fuse: FUSE

    @classmethod
    def configure_parser(cls, parser: argparse.ArgumentParser) -> None:
        super().configure_parser(parser)
        parser.add_argument(
            'mount',
            metavar='<mount>',
            action='store',
            help="Mount point",
        )

    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.mount = args.mount
        self.ino_map: Dict[str, int] = {}
        self.ino_count = 1

    def run(self) -> int:
        print(f"Mounting QOMFS to '{self.mount}'", file=sys.stderr)
        self.fuse = FUSE(self, self.mount, foreground=True)
        return 0

    def get_ino(self, path: str) -> int:
        """Get an inode number for a given QOM path."""
        if path in self.ino_map:
            return self.ino_map[path]
        self.ino_map[path] = self.ino_count
        self.ino_count += 1
        return self.ino_map[path]

    def is_object(self, path: str) -> bool:
        """Is the given QOM path an object?"""
        try:
            self.qom_list(path)
            return True
        except ExecuteError:
            return False

    def is_property(self, path: str) -> bool:
        """Is the given QOM path a property?"""
        path, prop = path.rsplit('/', 1)
        if path == '':
            path = '/'
        try:
            for item in self.qom_list(path):
                if item.name == prop:
                    return True
            return False
        except ExecuteError:
            return False

    def is_link(self, path: str) -> bool:
        """Is the given QOM path a link?"""
        path, prop = path.rsplit('/', 1)
        if path == '':
            path = '/'
        try:
            for item in self.qom_list(path):
                if item.name == prop and item.link:
                    return True
            return False
        except ExecuteError:
            return False

    def read(self, path: str, size: int, offset: int, fh: IO[bytes]) -> bytes:
        if not self.is_property(path):
            raise FuseOSError(ENOENT)

        path, prop = path.rsplit('/', 1)
        if path == '':
            path = '/'
        try:
            data = str(self.qmp.cmd('qom-get', path=path, property=prop))
            data += '\n'  # make values shell friendly
        except ExecuteError as err:
            raise FuseOSError(EPERM) from err

        if offset > len(data):
            return b''

        return bytes(data[offset:][:size], encoding='utf-8')

    def readlink(self, path: str) -> Union[bool, str]:
        if not self.is_link(path):
            return False
        path, prop = path.rsplit('/', 1)
        prefix = '/'.join(['..'] * (len(path.split('/')) - 1))
        return prefix + str(self.qmp.cmd('qom-get', path=path,
                                         property=prop))

    def getattr(self, path: str,
                fh: Optional[IO[bytes]] = None) -> Mapping[str, object]:
        if self.is_link(path):
            value = {
                'st_mode': 0o755 | stat.S_IFLNK,
                'st_ino': self.get_ino(path),
                'st_dev': 0,
                'st_nlink': 2,
                'st_uid': 1000,
                'st_gid': 1000,
                'st_size': 4096,
                'st_atime': 0,
                'st_mtime': 0,
                'st_ctime': 0
            }
        elif self.is_object(path):
            value = {
                'st_mode': 0o755 | stat.S_IFDIR,
                'st_ino': self.get_ino(path),
                'st_dev': 0,
                'st_nlink': 2,
                'st_uid': 1000,
                'st_gid': 1000,
                'st_size': 4096,
                'st_atime': 0,
                'st_mtime': 0,
                'st_ctime': 0
            }
        elif self.is_property(path):
            value = {
                'st_mode': 0o644 | stat.S_IFREG,
                'st_ino': self.get_ino(path),
                'st_dev': 0,
                'st_nlink': 1,
                'st_uid': 1000,
                'st_gid': 1000,
                'st_size': 4096,
                'st_atime': 0,
                'st_mtime': 0,
                'st_ctime': 0
            }
        else:
            raise FuseOSError(ENOENT)
        return value

    def readdir(self, path: str, fh: IO[bytes]) -> Iterator[str]:
        yield '.'
        yield '..'
        for item in self.qom_list(path):
            yield item.name



import threading
from functools import partial
from http.server import ThreadingHTTPServer, SimpleHTTPRequestHandler

class HttpServer:
    def __init__(self, listen:str, port:int, root:str):
        self.listen = listen
        self.port = port
        self.root = root

    def start(self):
        http_handler = partial(SimpleHTTPRequestHandler, directory=self.root)
        server = ThreadingHTTPServer((self.listen, self.port), http_handler)
        server_thread = threading.Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()



from typing import Optional
from datetime import datetime, timezone

def human_readable_size(num_bytes: int) -> str:
    units = ["B", "KB", "MB", "GB", "TB"]
    size = float(num_bytes)
    for unit in units:
        if size < 1024.0 or unit == units[-1]:
            return f"{size:.1f}{unit}"
        size /= 1024.0
    return f"{size:.1f}TB"

def humanize_age(created: Optional[datetime], now: Optional[datetime] = None) -> str:
    if created is None:
        return "<unknown>"
    base = now or datetime.now(timezone.utc)
    # Ensure timezone-aware comparison
    if created.tzinfo is None:
        created = created.replace(tzinfo=timezone.utc)
    delta = base - created

    seconds = int(delta.total_seconds())
    if seconds < 60: 
        return f"{seconds}s ago"
    minutes = seconds // 60
    if minutes < 60: 
        return f"{minutes}m ago"
    hours = minutes // 60
    if hours < 24: 
        return f"{hours}h ago"
    days = hours // 24
    if days < 30:
        return f"{days}d ago"
    months = days // 30
    if months < 12:
        return f"{months}mo ago"
    years = months // 12
    return f"{years}y ago"




from typing import Optional, List
import os
import sys
import tty
import re
import signal
import logging
import threading

from qemu_compose.utils.zio import zio, select_ignoring_useless_signal, write_debug, ttyraw
from qemu_compose.utils.jsonlisp import interp, default_env


logger = logging.getLogger("qemu-compose.instance.terminal")


class Terminal(object):
    def __init__(self, fd, log_path=None):
        self.fd = fd

        if isinstance(log_path, str):
            self.debug_file = open(log_path, "wb") if log_path else None
        else:
            self.debug_file = log_path

        self.io = zio(fd, print_write=False, logfile=sys.stdout, debug=self.debug_file, timeout=3600)

        self.term_feed_running = False
        self.term_feed_drain_thread = None

        if not os.isatty(0):
            raise Exception('qemu-compose.Terminal must run in a UNIX 98 style pty/tty')
        else:
            signal.signal(signal.SIGWINCH, self.handle_resize)

    def handle_resize(self, signum, frame):
        height, width = os.get_terminal_size(0)
        logger.info("try set terminal window size to %dx%d" % (width, height))
        # TODO: use qmp to set console window size

    def term_feed_loop(self):
        logger.info('Terminal.term_feed_loop started...')
        while self.term_feed_running:
            r, _, _ = select_ignoring_useless_signal([0], [], [], 0.2)

            if 0 in r:
                data = os.read(0, 1024)
                if data:
                    logger.info('Terminal.term_feed_loop received(%d) -> %s' % (len(data), data))
                    self.io.write(data)

        logger.info('Terminal.term_feed_loop finished.')

    def run_batch(self, cmds:List, env_variables=None):
        if not isinstance(cmds, list):
            raise ValueError("cmds must be a list")
        
        current_tty_mode = tty.tcgetattr(0)[:]
        ttyraw(0)

        try:
            self.term_feed_running = True
            self.term_feed_drain_thread = threading.Thread(target=self.term_feed_loop)
            self.term_feed_drain_thread.daemon = True
            self.term_feed_drain_thread.start()
            
            io = self.io

            if self.debug_file:
                write_debug(self.debug_file, b'run_batch: cmds = %r' % cmds)

            transpiled_cmds = ['begin'] + cmds

            env = default_env()

            env['read_until'] = io.read_until
            env['write'] = io.write
            env['writeline'] = io.writeline
            env['wait'] = io.read_until_timeout
            env['RegExp'] = lambda x: re.compile(x.encode())
            env['interact'] = self.interact

            if env_variables:
                env.update(env_variables)

            interp(transpiled_cmds, env)

        finally:
            tty.tcsetattr(0, tty.TCSAFLUSH, current_tty_mode)

    def interact(self, buffered:Optional[bytes]=None, raw_mode=False):

        self.term_feed_running = False
        if self.term_feed_drain_thread is not None:
            self.term_feed_drain_thread.join()

        self.io.interactive(raw_mode=raw_mode, buffered=buffered)



"""
(Legacy) Sync QMP Wrapper

This module provides the `QEMUMonitorProtocol` class, which is a
synchronous wrapper around `QMPClient`.

Its design closely resembles that of the original QEMUMonitorProtocol
class, originally written by Luiz Capitulino. It is provided here for
compatibility with scripts inside the QEMU source tree that expect the
old interface.
"""

#
# Copyright (C) 2009-2022 Red Hat Inc.
#
# Authors:
#  Luiz Capitulino <lcapitulino@redhat.com>
#  John Snow <jsnow@redhat.com>
#
# This work is licensed under the terms of the GNU GPL, version 2.  See
# the COPYING file in the top-level directory.
#

import asyncio
import socket
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from .error import QMPError
from .protocol import Runstate, SocketAddrT
from .qmp_client import QMPClient


#: QMPMessage is an entire QMP message of any kind.
QMPMessage = Dict[str, Any]

#: QMPReturnValue is the 'return' value of a command.
QMPReturnValue = object

#: QMPObject is any object in a QMP message.
QMPObject = Dict[str, object]

# QMPMessage can be outgoing commands or incoming events/returns.
# QMPReturnValue is usually a dict/json object, but due to QAPI's
# 'command-returns-exceptions', it can actually be anything.
#
# {'return': {}} is a QMPMessage,
# {} is the QMPReturnValue.


class QMPBadPortError(QMPError):
    """
    Unable to parse socket address: Port was non-numerical.
    """


class QEMUMonitorProtocol:
    """
    Provide an API to connect to QEMU via QEMU Monitor Protocol (QMP)
    and then allow to handle commands and events.

    :param address:  QEMU address, can be a unix socket path (string), a tuple
                     in the form ( address, port ) for a TCP connection, or an
                     existing `socket.socket` object.
    :param server:   Act as the socket server. (See 'accept')
                     Not applicable when passing a socket directly.
    :param nickname: Optional nickname used for logging.
    """

    def __init__(self,
                 address: Union[SocketAddrT, socket.socket],
                 server: bool = False,
                 nickname: Optional[str] = None):

        if server and isinstance(address, socket.socket):
            raise ValueError(
                "server argument should be False when passing a socket")

        self._qmp = QMPClient(nickname)
        self._aloop = asyncio.new_event_loop()
        self._address = address
        self._timeout: Optional[float] = None

        if server:
            assert not isinstance(self._address, socket.socket)
            self._sync(self._qmp.start_server(self._address))

    _T = TypeVar('_T')

    def _sync(
            self, future: Awaitable[_T], timeout: Optional[float] = None
    ) -> _T:
        return self._aloop.run_until_complete(
            asyncio.wait_for(future, timeout=timeout)
        )

    def _get_greeting(self) -> Optional[QMPMessage]:
        if self._qmp.greeting is not None:
            # pylint: disable=protected-access
            return self._qmp.greeting._asdict()
        return None

    def __enter__(self: _T) -> _T:
        # Implement context manager enter function.
        return self

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        # Implement context manager exit function.
        self.close()

    @classmethod
    def parse_address(cls, address: str) -> SocketAddrT:
        """
        Parse a string into a QMP address.

        Figure out if the argument is in the port:host form.
        If it's not, it's probably a file path.
        """
        components = address.split(':')
        if len(components) == 2:
            try:
                port = int(components[1])
            except ValueError:
                msg = f"Bad port: '{components[1]}' in '{address}'."
                raise QMPBadPortError(msg) from None
            return (components[0], port)

        # Treat as filepath.
        return address

    def connect(self, negotiate: bool = True) -> Optional[QMPMessage]:
        """
        Connect to the QMP Monitor and perform capabilities negotiation.

        :return: QMP greeting dict, or None if negotiate is false
        :raise ConnectError: on connection errors
        """
        self._qmp.await_greeting = negotiate
        self._qmp.negotiate = negotiate

        self._sync(
            self._qmp.connect(self._address)
        )
        return self._get_greeting()

    def accept(self, timeout: Optional[float] = 15.0) -> QMPMessage:
        """
        Await connection from QMP Monitor and perform capabilities negotiation.

        :param timeout:
            timeout in seconds (nonnegative float number, or None).
            If None, there is no timeout, and this may block forever.

        :return: QMP greeting dict
        :raise ConnectError: on connection errors
        """
        self._qmp.await_greeting = True
        self._qmp.negotiate = True

        self._sync(self._qmp.accept(), timeout)

        ret = self._get_greeting()
        assert ret is not None
        return ret

    def cmd_obj(self, qmp_cmd: QMPMessage) -> QMPMessage:
        """
        Send a QMP command to the QMP Monitor.

        :param qmp_cmd: QMP command to be sent as a Python dict
        :return: QMP response as a Python dict
        """
        return dict(
            self._sync(
                # pylint: disable=protected-access

                # _raw() isn't a public API, because turning off
                # automatic ID assignment is discouraged. For
                # compatibility with iotests *only*, do it anyway.
                self._qmp._raw(qmp_cmd, assign_id=False),
                self._timeout
            )
        )

    def cmd_raw(self, name: str,
                args: Optional[Dict[str, object]] = None) -> QMPMessage:
        """
        Build a QMP command and send it to the QMP Monitor.

        :param name: command name (string)
        :param args: command arguments (dict)
        """
        qmp_cmd: QMPMessage = {'execute': name}
        if args:
            qmp_cmd['arguments'] = args
        return self.cmd_obj(qmp_cmd)

    def cmd(self, cmd: str, **kwds: object) -> QMPReturnValue:
        """
        Build and send a QMP command to the monitor, report errors if any
        """
        return self._sync(
            self._qmp.execute(cmd, kwds),
            self._timeout
        )

    def pull_event(self,
                   wait: Union[bool, float] = False) -> Optional[QMPMessage]:
        """
        Pulls a single event.

        :param wait:
            If False or 0, do not wait. Return None if no events ready.
            If True, wait forever until the next event.
            Otherwise, wait for the specified number of seconds.

        :raise asyncio.TimeoutError:
            When a timeout is requested and the timeout period elapses.

        :return: The first available QMP event, or None.
        """
        if not wait:
            # wait is False/0: "do not wait, do not except."
            if self._qmp.events.empty():
                return None

        # If wait is 'True', wait forever. If wait is False/0, the events
        # queue must not be empty; but it still needs some real amount
        # of time to complete.
        timeout = None
        if wait and isinstance(wait, float):
            timeout = wait

        return dict(
            self._sync(
                self._qmp.events.get(),
                timeout
            )
        )

    def get_events(self, wait: Union[bool, float] = False) -> List[QMPMessage]:
        """
        Get a list of QMP events and clear all pending events.

        :param wait:
            If False or 0, do not wait. Return None if no events ready.
            If True, wait until we have at least one event.
            Otherwise, wait for up to the specified number of seconds for at
            least one event.

        :raise asyncio.TimeoutError:
            When a timeout is requested and the timeout period elapses.

        :return: A list of QMP events.
        """
        events = [dict(x) for x in self._qmp.events.clear()]
        if events:
            return events

        event = self.pull_event(wait)
        return [event] if event is not None else []

    def clear_events(self) -> None:
        """Clear current list of pending events."""
        self._qmp.events.clear()

    def close(self) -> None:
        """Close the connection."""
        self._sync(
            self._qmp.disconnect()
        )

    def settimeout(self, timeout: Optional[float]) -> None:
        """
        Set the timeout for QMP RPC execution.

        This timeout affects the `cmd`, `cmd_obj`, and `command` methods.
        The `accept`, `pull_event` and `get_event` methods have their
        own configurable timeouts.

        :param timeout:
            timeout in seconds, or None.
            None will wait indefinitely.
        """
        self._timeout = timeout

    def send_fd_scm(self, fd: int) -> None:
        """
        Send a file descriptor to the remote via SCM_RIGHTS.
        """
        self._qmp.send_fd_scm(fd)

    def __del__(self) -> None:
        if self._qmp.runstate == Runstate.IDLE:
            return

        if not self._aloop.is_running():
            self.close()
        else:
            # Garbage collection ran while the event loop was running.
            # Nothing we can do about it now, but if we don't raise our
            # own error, the user will be treated to a lot of traceback
            # they might not understand.
            raise QMPError(
                "QEMUMonitorProtocol.close()"
                " was not called before object was garbage collected"
            )




import datetime


'''
when handling datetime related, please follow rules below:

1. all datetime must set tzinfo
2. store and use datetime with tzinfo = UTC
3. only use local time when display to user interface
4. pass timestamp from epoch for API data
'''

beijing_timezone = datetime.timezone(datetime.timedelta(hours=8), "CST")

def to_timestamp(d:datetime.datetime):
    if not d.tzinfo:
        d = d.replace(tzinfo=datetime.timezone.utc)
    return d.timestamp()

def from_timestamp(d:float):
    return datetime.datetime.utcfromtimestamp(d).replace(tzinfo=datetime.timezone.utc)

def utcnow(with_tzinfo=True):
    d = datetime.datetime.utcnow()
    if with_tzinfo:
        return d.replace(tzinfo=datetime.timezone.utc)
    else:
        return d

def as_beijing_time(d:datetime.datetime):
    if not d.tzinfo:
        d = d.replace(tzinfo=datetime.timezone.utc)
    return d.astimezone(beijing_timezone)

def as_utc_time(d:datetime.datetime):
    if not d.tzinfo:
        d = d.replace(tzinfo=datetime.timezone.utc)
    return d.astimezone(datetime.timezone.utc)

def parse_datetime(v: str | int | float) -> datetime.datetime:
	if isinstance(v, (int, float)):
		return datetime.datetime.fromtimestamp(float(v), tz=datetime.timezone.utc)

	if not isinstance(v, str) or not v:
		return datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)

	s = v.strip()
	# Normalize trailing Z and trim fractional seconds to microseconds
	if s.endswith("Z"):
		s = s[:-1] + "+00:00"
	try:
		if "." in s:
			head, tail = s.split(".", 1)
			# tail may contain timezone like 190005011+00:00
			frac = ""
			tz = ""
			for i, ch in enumerate(tail):
				if ch.isdigit():
					frac += ch
				else:
					tz = tail[i:]
					break
			if tz == "" and len(frac) != len(tail):
				tz = tail[len(frac):]
			frac = (frac + "000000")[:6]
			s_norm = f"{head}.{frac}{tz}"
			return datetime.datetime.fromisoformat(s_norm)
		return datetime.datetime.fromisoformat(s)
	except Exception:
		# Fallback to epoch if parsing fails
		return datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)




import os

class LocalStore:
    def __init__(self, name="qemu-compose"):
        user_data_dir = os.getenv("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
        self.data_dir = os.path.join(user_data_dir, name)
        os.makedirs(self.data_dir, exist_ok=True)

    @property
    def image_root(self):
        path = os.path.join(self.data_dir, "image")
        os.makedirs(path, exist_ok=True)
        return path

    def image_dir(self, image_name):
        path = os.path.join(self.image_root, image_name)
        os.makedirs(path, exist_ok=True)
        return path

    @property
    def instance_root(self):
        path = os.path.join(self.data_dir, "instance")
        os.makedirs(path, exist_ok=True)
        return path
    
    def instance_dir(self, vmid):
        path = os.path.join(self.instance_root, vmid)
        os.makedirs(path, exist_ok=True)
        return path



from __future__ import annotations

from typing import Dict, Any
import random
import string

__all__ = ["generate_unique_name"]


RANDOM_NAME_ADJECTIVES = (
    "agile",
    "brisk",
    "calm",
    "daring",
    "eager",
    "fancy",
    "gentle",
    "happy",
    "jolly",
    "kind",
    "lively",
    "merry",
    "nimble",
    "proud",
    "quick",
    "ready",
    "smart",
    "tidy",
    "upbeat",
    "vivid",
)

RANDOM_NAME_NOUNS = (
    "badger",
    "beacon",
    "clover",
    "comet",
    "falcon",
    "feather",
    "harbor",
    "heron",
    "island",
    "jungle",
    "meadow",
    "nebula",
    "otter",
    "prairie",
    "quartz",
    "ranger",
    "spruce",
    "talon",
    "valley",
    "willow",
)


def generate_unique_name(existing_names: Dict[str, Any]) -> str:
    max_attempts = len(RANDOM_NAME_ADJECTIVES) * len(RANDOM_NAME_NOUNS)
    for _ in range(max_attempts):
        candidate = "-".join(
            (
                random.choice(RANDOM_NAME_ADJECTIVES),
                random.choice(RANDOM_NAME_NOUNS),
            )
        )
        if candidate not in existing_names:
            return candidate

    alphabet = string.ascii_lowercase + string.digits
    while True:
        suffix = "".join(random.choices(alphabet, k=4))
        candidate = "-".join(
            (
                random.choice(RANDOM_NAME_ADJECTIVES),
                random.choice(RANDOM_NAME_NOUNS),
                suffix,
            )
        )
        if candidate not in existing_names:
            return candidate



from typing import List, Optional
import datetime
import os
import json
from dataclasses import dataclass

from qemu_compose.utils.utcdatetime import parse_datetime

@dataclass(frozen=True)
class DiskSpec:
    filename: str
    format: str
    opts: str

    @classmethod
    def from_array(cls, a: List[str]) -> "DiskSpec":
        if len(a) == 0:
            return None
        fmt = a[1] if len(a) > 1 else "qcow2"
        opts = a[2] if len(a) > 2 else ""
        return DiskSpec(filename=a[0], format=fmt, opts=opts)

@dataclass(frozen=True)
class RepoTag:
    repo: str
    tag: str

    def match_name(self, name: str) -> bool:
        if ':' in name:
            r, t = name.split(':', maxsplit=1)
            return self.repo == r and self.tag == t
        else:
            return self.repo == name and self.tag == 'latest'

    @classmethod
    def from_str(cls, s:str) -> "RepoTag":
        if ':' in s:
            r, t = s.split(':', maxsplit=1)
            return RepoTag(repo=r, tag=t)
        else:
            return RepoTag(repo=s, tag='latest')

@dataclass(frozen=True)
class ImageManifest:
    id: str
    architecture: str
    os: str
    created: datetime.datetime
    repo_tags: List[RepoTag]
    disks: List[DiskSpec]
    qemu_args: List[str]
    digest: str
    comment: Optional[str]

    @classmethod
    def load_file(cls, image_dir: str) -> "ImageManifest":
        with open(os.path.join(image_dir, "manifest.json")) as f:
            obj = json.load(f)
        return cls.from_dict(obj)
    
    def has_repo_tag(self, name: str) -> bool:
        for rt in self.repo_tags:
            if rt.match_name(name):
                return True
        return False

    @classmethod
    def from_dict(cls, obj: dict) -> "ImageManifest":

        image_id = str(obj.get("id") or "")

        architecture = str(obj.get("architecture") or "")
        os_name = str(obj.get("os") or "")
        created = parse_datetime(obj.get("created"))

        repo_tags_raw = obj.get("repo_tags") or []
        repo_tags = [RepoTag.from_str(t) for t in repo_tags_raw if isinstance(t, str)]

        disks_raw = obj.get("disks") or []
        disks: List[DiskSpec] = []
        for item in disks_raw:
            if isinstance(item, list):
                ds = DiskSpec.from_array(item)
                if ds:
                    disks.append(ds)

        qemu_args_raw = obj.get("qemu_args") or []
        qemu_args = [str(a) for a in qemu_args_raw if isinstance(a, (str, int, float))]

        digest = str(obj.get("digest") or "")
        comment_val = obj.get("comment")
        comment = str(comment_val) if isinstance(comment_val, (str, int, float)) else None

        return cls(
            id=image_id,
            architecture=architecture,
            os=os_name,
            created=created,
            repo_tags=repo_tags,
            disks=disks,
            qemu_args=qemu_args,
            digest=digest,
            comment=comment,
        )



#!/usr/bin/env python3

import ast
import sys
import socket
from zio import zio

def parse(log_path:str):
    buf = [bytearray(), bytearray()]
    ret = []

    for line in open(log_path):
        if 'IO.recv(' in line:
            if buf[1]:
                ret.append((1, bytes(buf[1])))
                buf[1] = bytearray()
            i = line.index(' -> ')
            content = line[i+4:-1].strip()
            assert (content.startswith('b"') and content.endswith('"')) or (content.startswith("b'") and content.endswith("'"))
            buf[0].extend(ast.literal_eval(content))
        elif 'IO.send(' in line:
            if buf[0]:
                ret.append((0, bytes(buf[0])))
                buf[0] = bytearray()
            i = line.index(' -> ')
            content = line[i+4:-1].strip()
            assert (content.startswith('b"') and content.endswith('"')) or (content.startswith("b'") and content.endswith("'"))
            buf[1].extend(ast.literal_eval(content))
        else:
            continue

    if buf[0]:
        ret.append((0, bytes(buf[0])))
    if buf[1]:
        ret.append((1, bytes(buf[1])))
    return ret

def serve(logs, addr=None):
    server = socket.socket()
    if addr is None:
        addr = ('0.0.0.0', 1111)
    server.bind(addr)
    server.listen(1)
    s, _ = server.accept()
    io = zio(s, print_read=False, print_write=False, timeout=3600)
    for d, c in logs:
        if d == 1:
            print('EXPECT', c)
            io.read_until(c)
        else:
            print('SEND', c)
            io.write(c)
    io.close()

def main(name, log_path):
    logs = parse(log_path)
    if name == 'parse':
        for direction, content in logs:
            print('<<<<----' if direction else '---->>>>', content)
    elif name == 'serve':
        serve(logs)

if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2])



def to_valid_hostname(name: str) -> str:
    """Translate an arbitrary name to a valid Linux hostname label.

    Rules applied:
    - Keep only [a-z0-9-]; replace others with '-'
    - Lowercase the result
    - Collapse multiple '-'
    - Trim leading/trailing '-'
    - Truncate to 63 characters
    - If empty, fall back to 'vm'
    """
    import re

    s = name.lower()
    s = re.sub(r"[^a-z0-9-]", "-", s)
    s = re.sub(r"-+", "-", s)
    s = s.strip('-')
    if len(s) > 63:
        s = s[:63]
    if not s:
        s = 'vm'
    return s




from __future__ import annotations
from typing import Iterable, List, Tuple

from qemu_compose.local_store import LocalStore
from qemu_compose.image import list_image

def _print_table(rows: List[Tuple[str, str, str, str, str]]) -> None:
    headers = ["REPOSITORY", "TAG", "IMAGE ID", "CREATED", "SIZE"]

    def width(column: int) -> int:
        vals = [headers[column]] + [str(r[column]) for r in rows]
        return max(len(v) for v in vals)

    widths = [width(i) for i in range(len(headers))]

    def fmt_row(items: Iterable[str]) -> str:
        parts = [str(v).ljust(widths[i]) for i, v in enumerate(items)]
        return "  ".join(parts)

    print(fmt_row(headers))
    for r in rows:
        print(fmt_row(r))


def command_images() -> int:
    store = LocalStore()
    image_root = store.image_root
    rows = list_image(image_root)
    _print_table(rows)
    return 0




# NOTE: the following code comes from https://github.com/Leden/jsonlisp with small modification for this project
# but I have no idea why the original repo is 404 now, so I just leave a comment here to express thanks to Leden

import json
import logging
import operator
from collections import ChainMap
import ast

logger = logging.getLogger(__name__)

Symbol = str
List = list
Number = (int, float)
Dict = dict


class Env(ChainMap):
    def __json__(self):
        return self.maps


class Proc:
    def __init__(self, params, body, env):
        self.params = params
        self.body = body
        self.env = env

    def __call__(self, *args):
        env = Env(dict(zip(self.params, args)), self.env)
        return interp(self.body, env)

    def __json__(self):
        return ["lambda", self.params, self.body]


class Macro:
    def __init__(self, params, body, env):
        self.params = params
        self.body = body
        self.env = env

    def expand(self, *args, env):
        static_env = Env(dict(zip(self.params, args)), self.env)
        exp = interp(self.body, static_env)
        return interp(exp, env)

    def __json__(self):
        return ["macro", self.params, self.body]


builtins = {
    # Operators
    "*": operator.mul,
    "+": operator.add,
    "-": operator.sub,
    "/": operator.truediv,
    "<": operator.lt,
    "<=": operator.le,
    "=": operator.eq,
    ">": operator.gt,
    ">=": operator.ge,
    "^": operator.xor,
    "and": operator.and_,
    "contains": operator.contains,
    "in": lambda x, y: x in y,
    "is": operator.is_,
    "is-not": operator.is_not,
    "not": operator.not_,
    "or": operator.or_,
    "xor": lambda x, y: bool(x) ^ bool(y),  # logical XOR
    # Typechecks
    "dict?": lambda x: isinstance(x, Dict),
    "list?": lambda x: isinstance(x, List),
    "macro?": lambda x: isinstance(x, Macro),
    "null?": lambda x: x is None,
    "number?": lambda x: isinstance(x, Number),
    "proc?": callable,
    "symbol?": lambda x: isinstance(x, Symbol),
    # List functions
    "begin": lambda *x: x[-1],
    "cons": lambda x, y: [x, *y],
    "head": lambda x: x[0],
    "len": len,
    "list": lambda *x: list(x),
    "map": lambda *args: list(map(*args)),
    "range": lambda *x: list(range(*x)),
    "tail": lambda x: x[1:],
    # Dict functions
    "dict": lambda x: dict(x),
    "dict-del": lambda d, k: {_k: d[_k] for _k in d if k != _k},
    "dict-get": lambda d, k: d.get(k),
    "dict-items": lambda x: list(x.items()),
    "dict-set": lambda d, k, v: {**d, k: v},
    # Misc functions
    "apply": lambda proc, args: proc(*args),
    "print": print,
    "literal": ast.literal_eval,
    "str": str,
    "format": lambda x, *args: x % tuple(args),
}


std_lib = [
    "list",
    [
        "def",
        "defmacro",
        [
            "macro",
            ["name", "params", "body"],
            [
                "list",
                ["quote", "def"],
                "name",
                ["list", ["quote", "macro"], "params", "body"],
            ],
        ],
    ],
    [
        "defmacro",
        "defproc",
        ["name", "params", "body"],
        [
            "list",
            ["quote", "def"],
            "name",
            ["list", ["quote", "lambda"], "params", "body"],
        ],
    ],
]


def default_env():
    env = Env()
    env.update(builtins)
    env['key_up'] = "\x1b[A"
    env['key_down'] = "\x1b[B"
    env['key_right'] = "\x1b[C"
    env['key_left'] = "\x1b[D"
    env['key_home'] = "\x1b[H"
    env['key_end'] = "\x1b[F"
    env['key_ctrl_space'] = '\x00'
    env['key_escape'] = '\x1b'
    env['key_tab'] = '\t'
    env['key_enter'] = '\n'
    env['key_backspace'] = '\x7f'
    interp(std_lib, env)
    return env


def parse(line):
    return json.loads(line)


def parse_file(path):
    return json.load(path)


def unparse(expr):
    return json.dumps(expr, default=to_json)


def to_json(obj):
    if hasattr(obj, "__json__"):
        return obj.__json__()
    return repr(obj)


def interp(x, env):
    # a shortcut syntax for human readability
    if isinstance(x, Dict) and len(x) == 1:
        for k in x:
            v = x[k]
            if isinstance(v, List) or isinstance(v, Dict):
                v = interp(v, env)

            if not isinstance(v, List):
                v = [v]

            f = env[k]
            logger.info('CALL_STEP: %s %s' % (f.__name__, v))
            return f(*v)

    if isinstance(x, Symbol):
        if x.startswith('key_') and len(x) == 5:
            return x[4]
        return env[x]

    if not isinstance(x, List):
        return x

    if x == []:
        return x

    if x[0] == "quote" or x[0] == "'":
        _, exp = x
        return exp

    if x[0] == "flat_quote" or x[0] == "_'":
        _, *exp = x
        return exp

    if x[0] == "if":
        _, test, conseq, alt = x
        exp = conseq if interp(test, env) else alt
        return interp(exp, env)

    if x[0] == "def":
        _, var, exp = x
        val = env[var] = interp(exp, env)
        return val

    if x[0] == "lambda":
        _, params, body = x
        return Proc(params, body, env)

    if x[0] == "macro":
        _, params, body = x
        return Macro(params, body, env)

    proc_or_macro_exp, *args = x
    proc = interp(proc_or_macro_exp, env)
    if isinstance(proc, Macro):
        return proc.expand(*args, env=env)
    else:
        args = [interp(exp, env) for exp in args]
        if callable(proc):
            if proc_or_macro_exp != "begin":
                logger.info('CALL_STEP: %s(%s) %s' % (proc.__name__, proc_or_macro_exp, args))
        else:
            logger.error('CALL_STEP:ERROR: function not callable: %s(%s) %s' % (proc, proc_or_macro_exp, args))
        return proc(*args)

def repl(prompt=r"{λ}> "):
    env = default_env()

    while True:
        try:
            line = input(prompt)
            print(unparse(interp(parse(line), env)))
        except EOFError:
            print("Bye")
            break
        except Exception as err:
            print(repr(err))


def run_file(path):
    env = default_env()

    with open(path) as f:
        print(unparse(interp(parse_file(f), env)))



"""
QEMU Guest Agent Client

Usage:

Start QEMU with:

# qemu [...] -chardev socket,path=/tmp/qga.sock,server=on,wait=off,id=qga0 \
  -device virtio-serial \
  -device virtserialport,chardev=qga0,name=org.qemu.guest_agent.0

Run the script:

$ qemu-ga-client --address=/tmp/qga.sock <command> [args...]

or

$ export QGA_CLIENT_ADDRESS=/tmp/qga.sock
$ qemu-ga-client <command> [args...]

For example:

$ qemu-ga-client cat /etc/resolv.conf
# Generated by NetworkManager
nameserver 10.0.2.3
$ qemu-ga-client fsfreeze status
thawed
$ qemu-ga-client fsfreeze freeze
2 filesystems frozen

See also: https://wiki.qemu.org/Features/QAPI/GuestAgent
"""

# Copyright (C) 2012 Ryota Ozaki <ozaki.ryota@gmail.com>
#
# This work is licensed under the terms of the GNU GPL, version 2.  See
# the COPYING file in the top-level directory.

import argparse
import asyncio
import base64
import os
import random
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Sequence,
)

from qemu.qmp import ConnectError, SocketAddrT
from qemu.qmp.legacy import QEMUMonitorProtocol


# This script has not seen many patches or careful attention in quite
# some time. If you would like to improve it, please review the design
# carefully and add docstrings at that point in time. Until then:

# pylint: disable=missing-docstring


class QemuGuestAgent(QEMUMonitorProtocol):
    def __getattr__(self, name: str) -> Callable[..., Any]:
        def wrapper(**kwds: object) -> object:
            return self.cmd('guest-' + name.replace('_', '-'), **kwds)
        return wrapper


class QemuGuestAgentClient:
    def __init__(self, address: SocketAddrT):
        self.qga = QemuGuestAgent(address)
        self.qga.connect(negotiate=False)

    def sync(self, timeout: Optional[float] = 3) -> None:
        # Avoid being blocked forever
        if not self.ping(timeout):
            raise EnvironmentError('Agent seems not alive')
        uid = random.randint(0, (1 << 32) - 1)
        while True:
            ret = self.qga.sync(id=uid)
            if isinstance(ret, int) and int(ret) == uid:
                break

    def __file_read_all(self, handle: int) -> bytes:
        eof = False
        data = b''
        while not eof:
            ret = self.qga.file_read(handle=handle, count=1024)
            _data = base64.b64decode(ret['buf-b64'])
            data += _data
            eof = ret['eof']
        return data

    def read(self, path: str) -> bytes:
        handle = self.qga.file_open(path=path)
        try:
            data = self.__file_read_all(handle)
        finally:
            self.qga.file_close(handle=handle)
        return data

    def info(self) -> str:
        info = self.qga.info()

        msgs = []
        msgs.append('version: ' + info['version'])
        msgs.append('supported_commands:')
        enabled = [c['name'] for c in info['supported_commands']
                   if c['enabled']]
        msgs.append('\tenabled: ' + ', '.join(enabled))
        disabled = [c['name'] for c in info['supported_commands']
                    if not c['enabled']]
        msgs.append('\tdisabled: ' + ', '.join(disabled))

        return '\n'.join(msgs)

    @classmethod
    def __gen_ipv4_netmask(cls, prefixlen: int) -> str:
        mask = int('1' * prefixlen + '0' * (32 - prefixlen), 2)
        return '.'.join([str(mask >> 24),
                         str((mask >> 16) & 0xff),
                         str((mask >> 8) & 0xff),
                         str(mask & 0xff)])

    def ifconfig(self) -> str:
        nifs = self.qga.network_get_interfaces()

        msgs = []
        for nif in nifs:
            msgs.append(nif['name'] + ':')
            if 'ip-addresses' in nif:
                for ipaddr in nif['ip-addresses']:
                    if ipaddr['ip-address-type'] == 'ipv4':
                        addr = ipaddr['ip-address']
                        mask = self.__gen_ipv4_netmask(int(ipaddr['prefix']))
                        msgs.append(f"\tinet {addr}  netmask {mask}")
                    elif ipaddr['ip-address-type'] == 'ipv6':
                        addr = ipaddr['ip-address']
                        prefix = ipaddr['prefix']
                        msgs.append(f"\tinet6 {addr}  prefixlen {prefix}")
            if nif['hardware-address'] != '00:00:00:00:00:00':
                msgs.append("\tether " + nif['hardware-address'])

        return '\n'.join(msgs)

    def ping(self, timeout: Optional[float]) -> bool:
        self.qga.settimeout(timeout)
        try:
            self.qga.ping()
        except asyncio.TimeoutError:
            return False
        return True

    def fsfreeze(self, cmd: str) -> object:
        if cmd not in ['status', 'freeze', 'thaw']:
            raise ValueError('Invalid command: ' + cmd)
        # Can be int (freeze, thaw) or GuestFsfreezeStatus (status)
        return getattr(self.qga, 'fsfreeze' + '_' + cmd)()

    def fstrim(self, minimum: int) -> Dict[str, object]:
        # returns GuestFilesystemTrimResponse
        ret = getattr(self.qga, 'fstrim')(minimum=minimum)
        assert isinstance(ret, dict)
        return ret

    def suspend(self, mode: str) -> None:
        if mode not in ['disk', 'ram', 'hybrid']:
            raise ValueError('Invalid mode: ' + mode)

        try:
            getattr(self.qga, 'suspend' + '_' + mode)()
            # On error exception will raise
        except asyncio.TimeoutError:
            # On success command will timed out
            pass

    def shutdown(self, mode: str = 'powerdown') -> None:
        if mode not in ['powerdown', 'halt', 'reboot']:
            raise ValueError('Invalid mode: ' + mode)

        try:
            self.qga.shutdown(mode=mode)
        except asyncio.TimeoutError:
            pass


def _cmd_cat(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    if len(args) != 1:
        print('Invalid argument')
        print('Usage: cat <file>')
        sys.exit(1)
    print(client.read(args[0]))


def _cmd_fsfreeze(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    usage = 'Usage: fsfreeze status|freeze|thaw'
    if len(args) != 1:
        print('Invalid argument')
        print(usage)
        sys.exit(1)
    if args[0] not in ['status', 'freeze', 'thaw']:
        print('Invalid command: ' + args[0])
        print(usage)
        sys.exit(1)
    cmd = args[0]
    ret = client.fsfreeze(cmd)
    if cmd == 'status':
        print(ret)
        return

    assert isinstance(ret, int)
    verb = 'frozen' if cmd == 'freeze' else 'thawed'
    print(f"{ret:d} filesystems {verb}")


def _cmd_fstrim(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    if len(args) == 0:
        minimum = 0
    else:
        minimum = int(args[0])
    print(client.fstrim(minimum))


def _cmd_ifconfig(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    assert not args
    print(client.ifconfig())


def _cmd_info(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    assert not args
    print(client.info())


def _cmd_ping(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    timeout = 3.0 if len(args) == 0 else float(args[0])
    alive = client.ping(timeout)
    if not alive:
        print("Not responded in %s sec" % args[0])
        sys.exit(1)


def _cmd_suspend(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    usage = 'Usage: suspend disk|ram|hybrid'
    if len(args) != 1:
        print('Less argument')
        print(usage)
        sys.exit(1)
    if args[0] not in ['disk', 'ram', 'hybrid']:
        print('Invalid command: ' + args[0])
        print(usage)
        sys.exit(1)
    client.suspend(args[0])


def _cmd_shutdown(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    assert not args
    client.shutdown()


_cmd_powerdown = _cmd_shutdown


def _cmd_halt(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    assert not args
    client.shutdown('halt')


def _cmd_reboot(client: QemuGuestAgentClient, args: Sequence[str]) -> None:
    assert not args
    client.shutdown('reboot')


commands = [m.replace('_cmd_', '') for m in dir() if '_cmd_' in m]


def send_command(address: str, cmd: str, args: Sequence[str]) -> None:
    if not os.path.exists(address):
        print(f"'{address}' not found. (Is QEMU running?)")
        sys.exit(1)

    if cmd not in commands:
        print('Invalid command: ' + cmd)
        print('Available commands: ' + ', '.join(commands))
        sys.exit(1)

    try:
        client = QemuGuestAgentClient(address)
    except ConnectError as err:
        print(err)
        if isinstance(err.exc, ConnectionError):
            print('(Is QEMU running?)')
        sys.exit(1)

    if cmd == 'fsfreeze' and args[0] == 'freeze':
        client.sync(60)
    elif cmd != 'ping':
        client.sync()

    globals()['_cmd_' + cmd](client, args)


def main() -> None:
    address = os.environ.get('QGA_CLIENT_ADDRESS')

    parser = argparse.ArgumentParser()
    parser.add_argument('--address', action='store',
                        default=address,
                        help='Specify a ip:port pair or a unix socket path')
    parser.add_argument('command', choices=commands)
    parser.add_argument('args', nargs='*')

    args = parser.parse_args()
    if args.address is None:
        parser.error('address is not specified')
        sys.exit(1)

    send_command(args.address, args.command, args.args)


if __name__ == '__main__':
    main()
